@@ -1,6 +1,6 @@
 /*
  * ProFTPD: mod_facts -- a module for handling "facts" [RFC3659]
- * Copyright (c) 2007-2022 The ProFTPD Project
+ * Copyright (c) 2007-2025 The ProFTPD Project
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -460,7 +460,7 @@ static void facts_mlinfobuf_init(void) {
 static int facts_mlinfobuf_add(struct mlinfo *info, int flags) {
   char buf[PR_TUNABLE_BUFFER_SIZE];
   size_t buflen;
- 
+
   buflen = facts_mlinfo_fmt(info, buf, sizeof(buf), flags);
 
   /* If this buffer will exceed the capacity of mlinfo_buf, then flush
@@ -645,7 +645,7 @@ static int facts_mlinfo_get(struct mlinfo *info, const char *path,
             targetlen = pr_fsio_readlink(path, target, sizeof(target)-1);
           }
 
-          if (targetlen < 0) { 
+          if (targetlen < 0) {
             int xerrno = errno;
 
             pr_log_debug(DEBUG4, MOD_FACTS_VERSION
@@ -1533,7 +1533,7 @@ MODRET facts_mlsd(cmd_rec *cmd) {
 
   fake_mode = get_param_ptr(get_dir_ctxt(cmd->tmp_pool, (char *) best_path),
     "DirFakeMode", FALSE);
- 
+
   c = find_config(get_dir_ctxt(cmd->tmp_pool, (char *) best_path), CONF_PARAM,
     "DirFakeUser", FALSE);
   if (c != NULL) {
@@ -1621,9 +1621,9 @@ MODRET facts_mlsd(cmd_rec *cmd) {
 
     /* Check that the file can be listed. */
     abs_path = dir_realpath(cmd->tmp_pool, rel_path);
-    if (abs_path) {
+    if (abs_path != NULL) {
       res = dir_check(cmd->tmp_pool, cmd, cmd->group, abs_path, &hidden);
-      
+
     } else {
       abs_path = dir_canonical_path(cmd->tmp_pool, rel_path);
       if (abs_path == NULL) {
@@ -1994,19 +1994,20 @@ MODRET facts_opts_mlst(cmd_rec *cmd) {
 
 /* usage: FactsAdvertise on|off */
 MODRET set_factsadvertise(cmd_rec *cmd) {
-  int bool = -1;
+  int advertise_facts = -1;
   config_rec *c = NULL;
 
   CHECK_ARGS(cmd, 1);
   CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
 
-  bool = get_boolean(cmd, 1);
-  if (bool == -1)
+  advertise_facts = get_boolean(cmd, 1);
+  if (advertise_facts == -1) {
     CONF_ERROR(cmd, "expected Boolean parameter");
+  }
 
   c = add_config_param(cmd->argv[0], 1, NULL);
   c->argv[0] = pcalloc(c->pool, sizeof(int));
-  *((int *) c->argv[0]) = bool;
+  *((int *) c->argv[0]) = advertise_facts;
 
   return PR_HANDLED(cmd);
 }
@@ -2079,6 +2080,13 @@ MODRET set_factsoptions(cmd_rec *cmd) {
   c->argv[0] = palloc(c->pool, sizeof(unsigned long));
   *((unsigned long *) c->argv[0]) = opts;
 
+  if (pr_module_exists("mod_ifsession.c")) {
+    /* These are needed in case this directive is used with mod_ifsession
+     * configuration.
+     */
+    c->flags |= CF_MULTI;
+  }
+
   return PR_HANDLED(cmd);
 }
 
@@ -2146,7 +2154,7 @@ static int facts_sess_init(void) {
     unsigned long opts;
 
     pr_signals_handle();
-  
+
     opts = *((unsigned long *) c->argv[0]);
     facts_mlinfo_opts |= opts;
 
